#include "fresh.h"

#include <Arduino.h>
#include <time.h>
#include <EEPROM.h>

/**
 * @brief The size of the message counter in bytes (uint32_t)
 */
#define MESSAGE_COUNTER_SIZE sizeof(uint32_t)

/**
 * @brief The allowed discrepancy between the time of a received message
 * and the device time (in seconds)
 * 
 * A stricter (lower) value better prevents against replay attacks,
 * but may lead to issues when dealing with slow networks and other
 * routing delays.
 */
uint32_t allowedOffset = 60;

void setMessageTimeAllowedOffset(uint32_t offset) {
    allowedOffset = offset;
}

void configureNTP(int32_t offsetToGMT, int32_t offsetDaylightSavings, const char* serverUrl) {
    configTime(offsetToGMT, offsetDaylightSavings, serverUrl);
}

void printLocalTime() {
    struct tm timeinfo;
    if (getLocalTime(&timeinfo)) {
        Serial.println(&timeinfo, "[INFO] Time is %A, %d. %B %Y %H:%M:%S");
    } else {
        Serial.println("[WARN] No local time available");
    }
}

uint32_t getEpochTime() {
    time_t now;
    struct tm timeinfo;
    if (!getLocalTime(&timeinfo)) {
        Serial.println("[WARN] Failed to obtain local time");
        return(0);
    }
    time(&now);
    return now;
}

bool isMessageTimeAcceptable(uint32_t t) {
    uint32_t localTime = getEpochTime();
    if (localTime == 0) {
        return false;
    }
    return t < localTime + allowedOffset && t > localTime - allowedOffset;
}

void prepareMessageCounterUsage() {
    EEPROM.begin(MESSAGE_COUNTER_SIZE);
}

uint32_t getNextMessageCounter() {
    uint32_t counter = (uint32_t) EEPROM.read(0) << 24;
    counter += (uint32_t) EEPROM.read(1) << 16;
    counter += (uint32_t) EEPROM.read(2) << 8;
    counter += (uint32_t) EEPROM.read(3);
    return counter;
}

void printMessageCounter() {
    Serial.printf("[INFO] Next message number: %d\n", getNextMessageCounter());
}

bool isMessageCounterValid(uint32_t counter) {
    return counter >= getNextMessageCounter();
}

void didUseMessageCounter(uint32_t counter) {
    // Store the next counter, so that resetting starts at 0
    counter += 1;
    EEPROM.write(0, (counter >> 24) & 0xFF);
    EEPROM.write(1, (counter >> 16) & 0xFF);
    EEPROM.write(2, (counter >>  8) & 0xFF);
    EEPROM.write(3,  counter        & 0xFF);

    EEPROM.commit();
}

void resetMessageCounter() {
    didUseMessageCounter(0);
    Serial.println("[WARN] Message counter reset");
}
