/**
 * Sesame-Device
 * Christoph Hagen, 2022.
 *
 * The code for a simple door unlock mechanism where a servo pushes on an existing
 * physical button.
 */
#include <Arduino.h>
#include <WiFi.h>

#include "crypto.h"
#include "fresh.h"
#include "message.h"
#include "server.h"
#include "servo.h"

// TODO: 
// - Handle WiFi disconnect
// - Implement key sharing
// - Configure/diagnose over Bluetooth?

/* Settings */

constexpr uint32_t serialBaudRate = 115200;

constexpr size_t keySize = 32;
constexpr const uint8_t remoteKey[keySize] = { 1, 2, 3};
constexpr const uint8_t localKey[keySize] = { 1, 2, 3};

// The WiFi network to connect to
constexpr const char* wifiSSID = "MyNetwork";
constexpr const char* wifiPassword = "MyPassword";
constexpr uint32_t wifiReconnectInterval = 10000;

// The remote server to connect to
constexpr const char* serverUrl = "mydomain.com";
constexpr const int serverPort = 443;
constexpr const char* serverPath = "/sesame/listen";
constexpr const char* serverAccessKey = "MyAccessToken";

/* Time */
constexpr const char* ntpServerUrl = "pool.ntp.org";
constexpr int32_t timeOffsetToGMT = 3600;
constexpr int32_t timeOffsetDaylightSavings = 3600;

/* Servo */

// Servo is Emax ES08MA II

// The time (in ms) to keep the door button pressed
constexpr uint32_t lockOpeningDuration = 2000;

// The timer to use to control the servo
constexpr int pwmTimer = 0;

// Pin wired up to the servo data line
constexpr int servoPin = 14;

// The Emax is a standard 50 Hz servo
constexpr int servoFrequency = 50;

// The microseconds to set the servo to the pressed and released states
constexpr int servoPressedState = 1600;
constexpr int servoReleasedState = 1520;


/* Global variables */

ServerConnection server(serverUrl, serverPort, serverPath);

ServoController servo(pwmTimer, servoFrequency, servoPin);

/* Event callbacks */

SesameEvent handleReceivedMessage(AuthenticatedMessage* payload, AuthenticatedMessage* response);

/* Logic */

void setup() {
    Serial.begin(serialBaudRate);
    Serial.setDebugOutput(true);
    Serial.println("[INFO] Device started");

    servo.configure(lockOpeningDuration, servoPressedState, servoReleasedState);
    Serial.println("[INFO] Servo configured");

    prepareMessageCounterUsage();
    printMessageCounter();
    
    Serial.printf("[INFO] Connecting to WiFi '%s'\n", wifiSSID);
    WiFi.begin(wifiSSID, wifiPassword);
    while(WiFi.status() != WL_CONNECTED) {
        delay(100);
    }
    Serial.println("[INFO] WiFi connected");

    configureNTP(timeOffsetToGMT, timeOffsetDaylightSavings, ntpServerUrl);
    printLocalTime();
    
    server.onMessage(handleReceivedMessage);
    Serial.printf("[INFO] Opening SSL socket %s%s on port %d\n", serverUrl, serverPath, serverPort);
    server.connectSSL(serverAccessKey);
}

void loop() {
    server.loop();
    servo.loop();
}

/**
 * Process received binary data.
 *
 * Checks whether the received data is a valid and unused key,
 * and then signals that the motor should move.
 *
 * @param payload The pointer to the received data.
 * @param length The number of bytes received.
 * @return The event to signal to the server.
 */
SesameEvent handleReceivedMessage(AuthenticatedMessage* message, AuthenticatedMessage* response) {
    
    if (!isMessageCounterValid(message->message.id)) {
        return SesameEvent::MessageCounterInvalid;
    }
    if (!isMessageTimeAcceptable(message->message.time)) {
        return SesameEvent::MessageTimeMismatch;
    }
    if (!isAuthenticMessage(message, remoteKey, keySize)) {
        return SesameEvent::MessageAuthenticationFailed;
    }

    response->message.time = getEpochTime();
    response->message.id = getNextMessageCounter();
    if (!authenticateMessage(response, localKey, keySize)) {
        return SesameEvent::MessageAuthenticationFailed;
    }

    // Move servo
    servo.pressButton();
    Serial.printf("[Info] Accepted message %d\n", message->message.id);
    return SesameEvent::MessageAccepted;
}
